@@ -2,19 +2,23 @@
  * 
  * This file is part of JSON-GLib
  * Copyright (C) 2007  OpenedHand Ltd.
+ * Copyright (C) 2009  Intel Corp.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library. If not, see <http://www.gnu.org/licenses/>.
+ *
  * Author:
- *   Emmanuele Bassi  <ebassi@openedhand.com>
+ *   Emmanuele Bassi  <ebassi@linux.intel.com>
  */
 
 /**
@@ -32,6 +36,8 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include "json-types-private.h"
+
 #include "json-marshal.h"
 #include "json-generator.h"
 
@@ -73,8 +79,39 @@ static gchar *dump_object (JsonGenerator *generator,
                            JsonObject    *object,
                            gsize         *length);
 
+/* non-ASCII characters can't be escaped, otherwise UTF-8
+ * chars will break, so we just pregenerate this table of
+ * high characters and then we feed it to g_strescape()
+ */
+static const char json_exceptions[] = {
+  0x7f,  0x80,  0x81,  0x82,  0x83,  0x84,  0x85,  0x86,
+  0x87,  0x88,  0x89,  0x8a,  0x8b,  0x8c,  0x8d,  0x8e,
+  0x8f,  0x90,  0x91,  0x92,  0x93,  0x94,  0x95,  0x96,
+  0x97,  0x98,  0x99,  0x9a,  0x9b,  0x9c,  0x9d,  0x9e,
+  0x9f,  0xa0,  0xa1,  0xa2,  0xa3,  0xa4,  0xa5,  0xa6,
+  0xa7,  0xa8,  0xa9,  0xaa,  0xab,  0xac,  0xad,  0xae,
+  0xaf,  0xb0,  0xb1,  0xb2,  0xb3,  0xb4,  0xb5,  0xb6,
+  0xb7,  0xb8,  0xb9,  0xba,  0xbb,  0xbc,  0xbd,  0xbe,
+  0xbf,  0xc0,  0xc1,  0xc2,  0xc3,  0xc4,  0xc5,  0xc6,
+  0xc7,  0xc8,  0xc9,  0xca,  0xcb,  0xcc,  0xcd,  0xce,
+  0xcf,  0xd0,  0xd1,  0xd2,  0xd3,  0xd4,  0xd5,  0xd6,
+  0xd7,  0xd8,  0xd9,  0xda,  0xdb,  0xdc,  0xdd,  0xde,
+  0xdf,  0xe0,  0xe1,  0xe2,  0xe3,  0xe4,  0xe5,  0xe6,
+  0xe7,  0xe8,  0xe9,  0xea,  0xeb,  0xec,  0xed,  0xee,
+  0xef,  0xf0,  0xf1,  0xf2,  0xf3,  0xf4,  0xf5,  0xf6,
+  0xf7,  0xf8,  0xf9,  0xfa,  0xfb,  0xfc,  0xfd,  0xfe,
+  0xff,
+  '\0'   /* g_strescape() expects a NUL-terminated string */
+};
+
 G_DEFINE_TYPE (JsonGenerator, json_generator, G_TYPE_OBJECT);
 
+static gchar *
+json_strescape (const gchar *str)
+{
+  return g_strescape (str, json_exceptions);
+}
+
 static void
 json_generator_finalize (GObject *gobject)
 {
@@ -235,12 +272,13 @@ dump_value (JsonGenerator *generator,
   guint indent = priv->indent;
   GValue value = { 0, };
   GString *buffer;
-  guint i;
 
   buffer = g_string_new ("");
 
   if (pretty)
     {
+      guint i;
+
       for (i = 0; i < (level * indent); i++)
         g_string_append_c (buffer, priv->indent_char);
     }
@@ -250,22 +288,30 @@ dump_value (JsonGenerator *generator,
 
   json_node_get_value (node, &value);
 
-
-  gchar double_str[20] = {0};
-  
   switch (G_VALUE_TYPE (&value))
     {
-    case G_TYPE_INT:
-      g_string_append_printf (buffer, "%d", g_value_get_int (&value));
+    case G_TYPE_INT64:
+      g_string_append_printf (buffer, "%" G_GINT64_FORMAT, g_value_get_int64 (&value));
       break;
 
     case G_TYPE_STRING:
-      g_string_append_printf (buffer, "\"%s\"", g_value_get_string (&value));
+      {
+        gchar *tmp;
+
+        tmp = json_strescape (g_value_get_string (&value));
+        g_string_append_printf (buffer, "\"%s\"", tmp);
+
+        g_free (tmp);
+      }
       break;
 
     case G_TYPE_DOUBLE:
-      g_ascii_formatd (double_str, 20, "%f", g_value_get_double (&value));
-      g_string_append (buffer, double_str);
+      {
+        gchar buf[65];
+
+        g_ascii_formatd (buf, 65, "%g", g_value_get_double (&value));
+        g_string_append (buffer, buf);
+      }
       break;
 
     case G_TYPE_BOOLEAN:
@@ -491,13 +537,14 @@ json_generator_new (void)
 /**
  * json_generator_to_data:
  * @generator: a #JsonGenerator
- * @length: return location for the length of the returned buffer, or %NULL
+ * @length: (out): return location for the length of the returned
+ *   buffer, or %NULL
  *
  * Generates a JSON data stream from @generator and returns it as a
  * buffer.
  *
- * Return value: a newly allocated buffer holding a JSON data stream. Use
- *   g_free() to free the allocated resources.
+ * Return value: a newly allocated buffer holding a JSON data stream.
+ *   Use g_free() to free the allocated resources.
  */
 gchar *
 json_generator_to_data (JsonGenerator *generator,
@@ -571,6 +618,43 @@ json_generator_to_file (JsonGenerator  *generator,
   return retval;
 }
 
+/**
+ * json_generator_to_stream:
+ * @generator: a #JsonGenerator
+ * @stream: a #GOutputStream
+ * @cancellable: (allow-none): a #GCancellable, or %NULL
+ * @error: return location for a #GError, or %NULL
+ *
+ * Outputs JSON data and streams it (synchronously) to @stream.
+ *
+ * Return value: %TRUE if the write operation was successful, and %FALSE
+ *   on failure. In case of error, the #GError will be filled accordingly
+ *
+ * Since: 0.12
+ */
+gboolean
+json_generator_to_stream (JsonGenerator  *generator,
+                          GOutputStream  *stream,
+                          GCancellable   *cancellable,
+                          GError        **error)
+{
+  gboolean retval;
+  gchar *buffer;
+  gsize len;
+
+  g_return_val_if_fail (JSON_IS_GENERATOR (generator), FALSE);
+  g_return_val_if_fail (G_IS_OUTPUT_STREAM (stream), FALSE);
+
+  if (g_cancellable_set_error_if_cancelled (cancellable, error))
+    return FALSE;
+
+  buffer = json_generator_to_data (generator, &len);
+  retval = g_output_stream_write (stream, buffer, len, cancellable, error);
+  g_free (buffer);
+
+  return retval;
+}
+
 /**
  * json_generator_set_root:
  * @generator: a #JsonGenerator
@@ -579,8 +663,8 @@ json_generator_to_file (JsonGenerator  *generator,
  * Sets @node as the root of the JSON data stream to be serialized by
  * the #JsonGenerator.
  *
- * Note: the node is copied by the generator object, so it can be safely
- * freed after calling this function.
+ * <note>The node is copied by the generator object, so it can be safely
+ * freed after calling this function.</note>
  */
 void
 json_generator_set_root (JsonGenerator *generator,
